  (*
   *
   *  Copyright (c) 2005, 2006 Reynald Affeldt and Nicolas Marti
   *
   *  This program is free software; you can redistribute it and/or modify
   *  it under the terms of the GNU General Public License as published by
   *  the Free Software Foundation; either version 2 of the License, or
   *  (at your option) any later version.
   *
   *  This program is distributed in the hope that it will be useful,
   *  but WITHOUT ANY WARRANTY; without even the implied warranty of
   *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   *  GNU General Public License for more details.
   *
   *  You should have received a copy of the GNU General Public License
   *  along with this program; if not, write to the Free Software
   *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
   *)

(*
  

  http://users.rsise.anu.edu.au/~michaeln/pubs/arith-dp-slides.pdf
  http://www.cse.unsw.edu.au/~cs4161/michaeln-lecture14.pdf
  http://users.rsise.anu.edu.au/~michaeln/pubs/omega-slides.pdf
  http://cs156.stanford.edu/notes/dp-4.pdf
*)


Require Import ZArith.
Require Import List.
Require Import EqNat.

Require Import util.
Require Import util_tactic.

Require Import Classical.


(* Environment for arithmetic and boolean deep embedding 
  
  The value a the variable i is the ith element of its environment

  The goal of this decision procedure: prove that for every environment a system of arithmetic constraint is true.

  ==Z> We try to eliminate all the variables, such that to evaluate the system we do not need an environment (and thus this evaluation is the same for every environment)

  Note: in fact we negate the formula of the system and find if for every environment its evaluation is false (Came from the fact that 

  ~ forall x, P x <-> exists x, ~ P x

)



*)

Definition environmentZ := list Z.

(* same as In of the standard library but returning a boolean *)

Fixpoint inb (A: Set) (B: A -> A -> bool) (l: list A) (a: A) {struct l}: bool :=
  match l with
    nil => false
    | hd::tl => if (B a hd) then true else (inb A B tl a)
  end.
Implicit Arguments inb.

(* add an element to a list only if not already include *)

Fixpoint add_elt_list (A: Set) (B: A -> A -> bool) (l: list A) (a: A) {struct l}: list A :=
  match l with 
    nil => a::nil
    | hd::tl => if (B hd a) then l else hd::(add_elt_list A B tl a)
  end.
Implicit Arguments add_elt_list.

(* append two lists using the previous functions (adding only the elements not already include) *)

Fixpoint app_list (A: Set) (B: A -> A -> bool) (l1 l2: list A){struct l1}: list A :=
  match l1 with 
    nil => l2
    | hd::tl => app_list A B tl (add_elt_list B l2 hd)
  end.
Implicit Arguments app_list.

  (***********************************)

(* append two option lists *)

Definition option_app (A: Set) (l1 l2: option (list A)) : option (list A) :=
  match l1 with
    None => None
    | Some l1' =>   match l2 with
                      None => None
                      | Some l2' => Some (l1' ++ l2')
                    end
  end.
Implicit Arguments option_app.

(* andb for option bool *)

Definition andb_option (a b: option bool) :=
  match a with
    None => None
    | Some a' => match b with
                           None => None
                           | Some b' => Some (andb a' b')
                         end
  end.

(* orb for option bool *)

Definition orb_option (a b: option bool) :=
  match a with
    None => None
    | Some a' => match b with
                           None => None
                           | Some b' => Some (orb a' b')
                         end
  end.

  (***********************************)


(* Deep encoding of arithmetic expressions *)

Inductive exprZ : Set :=
  var_eZ : nat -> exprZ
  | int_eZ : Z -> exprZ
  | plus_eZ : exprZ -> exprZ -> exprZ
  | minus_eZ : exprZ -> exprZ -> exprZ
  | mult_eZ : exprZ -> exprZ -> exprZ.

Notation "e1 '+eZ' e2" := (plus_eZ e1 e2) (at level 79) : presburgerZ_scope.
Notation "e1 '-eZ' e2" := (minus_eZ e1 e2) (at level 79) : presburgerZ_scope.
Notation "e1 '*eZ' e2" := (mult_eZ e1 e2) (at level 79) : presburgerZ_scope.

(* Evaluation of arithmetic expressions using an environment *)

Fixpoint eval_exprZ (e: exprZ) (env: environmentZ) {struct e} : Z :=
  match e with
    var_eZ x => nth x env 0%Z
    | int_eZ y => y
    | plus_eZ e1 e2 => (eval_exprZ e1 env + eval_exprZ e2 env)%Z
    | minus_eZ e1 e2 => (eval_exprZ e1 env - eval_exprZ e2 env)%Z
    | mult_eZ e1 e2 => (eval_exprZ e1 env * eval_exprZ e2 env)%Z
  end.

(* Variable of arithmetic expressions *)

Fixpoint exprZ_var (e: exprZ) : list nat :=
  match e with
    var_eZ x => x::nil
    | int_eZ y => nil
    | plus_eZ e1 e2 => exprZ_var e1 ++ exprZ_var e2
    | minus_eZ e1 e2 => exprZ_var e1 ++ exprZ_var e2 
    | mult_eZ e1 e2 => exprZ_var e1 ++ exprZ_var e2
  end.


Open Local Scope presburgerZ_scope.

(* evaluation without environment (None if cannot be evaluated) *)

Fixpoint compute_exprZ (e: exprZ) : option Z :=
  match e with
    var_eZ x => None
    | int_eZ x => Some x
    | e1 +eZ e2 => match (compute_exprZ e1) with
                    None => None
                    | Some e1' => match (compute_exprZ e2) with
                                    None => None
                                    | Some e2' => Some (e1' + e2')%Z
                                  end
                  end
    | e1 -eZ e2 => match (compute_exprZ e1) with
                    None => None
                    | Some e1' => match (compute_exprZ e2) with
                                    None => None
                                    | Some e2' => Some (e1' - e2')%Z
                                  end
                  end
    | e1 *eZ e2 => match (compute_exprZ e1) with
                    None => match (compute_exprZ e2) with
                              None => None
                              | Some e2' => if (Zeq_bool e2' 0%Z) then (Some 0%Z) else None
                            end
                    | Some e1' => if (Zeq_bool e1' 0%Z) then (Some 0%Z) else 
                      match (compute_exprZ e2) with
                        None => None
                        | Some e2' => Some (e1' * e2')%Z
                      end
                  end
  end.

(* if an arithmetic expression can be evaluated without environment, this value is correct *)

Lemma compute_exprZ_correct: forall e z,
  compute_exprZ e = Some z ->
  (forall s, eval_exprZ e s = z).

  induction e; intros; simpl in H; try discriminate; simpl.
  injection H; intros; auto.
  destruct (compute_exprZ e1); destruct (compute_exprZ e2); simpl in H; try discriminate; injection H; intros; subst z.
  rewrite (IHe1 z0); auto; rewrite (IHe2 z1); auto.
  destruct (compute_exprZ e1); destruct (compute_exprZ e2); simpl in H; try discriminate; injection H; intros; subst z.
  rewrite (IHe1 z0); auto; rewrite (IHe2 z1); auto.
  destruct (compute_exprZ e1); destruct (compute_exprZ e2); simpl in H; try discriminate.
  generalize (Zeq_bool_classic z0 0); intro X; inversion_clear X; rewrite H0 in H; injection H; intros; subst z.
  rewrite (IHe1 z0); auto; rewrite (IHe2 z1); auto.
  rewrite (Zeq_bool_true _ _ H0); ring.
  rewrite (IHe1 z0); auto; rewrite (IHe2 z1); auto.
  generalize (Zeq_bool_classic z0 0); intro X; inversion_clear X; rewrite H0 in H; try ((injection H; intros; subst z) || inversion H).
  rewrite (IHe1 z0); auto.
  rewrite (Zeq_bool_true _ _ H0); ring.
  generalize (Zeq_bool_classic z0 0); intro X; inversion_clear X; rewrite H0 in H; try ((injection H; intros; subst z) || inversion H).
  rewrite (IHe2 z0); auto.
  rewrite (Zeq_bool_true _ _ H0); ring.
Qed.


(* Deep embedding of boolean expressions *)

Inductive exprBZ : Set :=
  eq_bZ: exprZ -> exprZ -> exprBZ
  | gt_bZ: exprZ -> exprZ -> exprBZ
  | ge_bZ: exprZ -> exprZ -> exprBZ
  | lt_bZ: exprZ -> exprZ -> exprBZ
  | le_bZ: exprZ -> exprZ -> exprBZ
  | or_bZ: exprBZ -> exprBZ -> exprBZ
  | and_bZ: exprBZ -> exprBZ -> exprBZ
  | neg_bZ: exprBZ -> exprBZ.

Notation "e ==Z e'" := (eq_bZ e e') (at level 78) : presburgerZ_scope.
Notation "e =/=Z e'" := (neg_bZ (eq_bZ e e')) (at level 78) : presburgerZ_scope.
Notation "e >>=Z e'" := (ge_bZ e e') (at level 78) : presburgerZ_scope.
Notation "e >>Z e'" := (gt_bZ e e') (at level 78) : presburgerZ_scope.
Notation "e <<Z e'" := (lt_bZ e e') (at level 78) : presburgerZ_scope.
Notation "e <<=Z e'" := (le_bZ e e') (at level 78) : presburgerZ_scope.
Notation "e &&&Z e'" := (and_bZ e e') (at level 78) : presburgerZ_scope.
Notation "e |||Z e'" := (or_bZ e e') (at level 78) : presburgerZ_scope.
Notation "!Z e" := (neg_bZ e) (at level 78) : presburgerZ_scope.
Notation " b1 =bZ> b2 " := ((!Z b1) |||Z b2) (at level 78) : presburgerZ_scope.

(* Evaluation of boolean expressions with an environment *)

Fixpoint eval_exprBZ (e:exprBZ) (env: environmentZ) {struct e} : bool  :=
  match e with 
    eq_bZ e1 e2 => Zeq_bool (eval_exprZ e1 env) (eval_exprZ e2 env)
    | gt_bZ e1 e2 => Zgt_bool (eval_exprZ e1 env) (eval_exprZ e2 env)
    | ge_bZ e1 e2 => Zge_bool (eval_exprZ e1 env) (eval_exprZ e2 env)
    | lt_bZ e1 e2 => Zlt_bool (eval_exprZ e1 env) (eval_exprZ e2 env)
    | le_bZ e1 e2 => Zle_bool (eval_exprZ e1 env) (eval_exprZ e2 env)
    | or_bZ e1 e2 => orb (eval_exprBZ e1 env) (eval_exprBZ e2 env)
    | and_bZ e1 e2 => andb (eval_exprBZ e1 env) (eval_exprBZ e2 env)
    | neg_bZ e => negb (eval_exprBZ e env)
  end.

Notation " s '|bZ=' b " := (eval_exprBZ b s = true) (at level 78) : presburgerZ_scope.
Notation " s '|b!Z=' b " := (eval_exprBZ b s = false) (at level 78) : presburgerZ_scope.

(* variables of boolean expressions *)

Fixpoint exprBZ_var (e:exprBZ) : list nat :=
  match e with 
    eq_bZ e1 e2 => exprZ_var e1 ++ exprZ_var e2
    | gt_bZ e1 e2 => exprZ_var e1 ++ exprZ_var e2
    | ge_bZ e1 e2 => exprZ_var e1 ++ exprZ_var e2
    | lt_bZ e1 e2 => exprZ_var e1 ++ exprZ_var e2
    | le_bZ e1 e2 => exprZ_var e1 ++ exprZ_var e2
    | or_bZ e1 e2 => exprBZ_var e1 ++ exprBZ_var e2
    | and_bZ e1 e2 => exprBZ_var e1 ++ exprBZ_var e2
    | neg_bZ e => exprBZ_var e
  end.

(* translation from of boolean expressions to Coq language *)


Fixpoint exprBZ_sem (e: exprBZ) (s: environmentZ) {struct e} : Prop :=
  match e with
    | f ==Z g => eval_exprZ f s = eval_exprZ g s
    | f >>=Z g => (eval_exprZ f s >= eval_exprZ g s)%Z
    | f >>Z g => (eval_exprZ f s > eval_exprZ g s)%Z
    | f <<=Z g => (eval_exprZ f s <= eval_exprZ g s)%Z
    | f <<Z g => (eval_exprZ f s < eval_exprZ g s)%Z
    | f &&&Z g => exprBZ_sem f s /\ exprBZ_sem g s
    | f |||Z g => exprBZ_sem f s \/ exprBZ_sem g s
    | neg_bZ e => ~ exprBZ_sem e s
  end. 

(* correctness of the translation *)

Lemma exprBZ_semantic_good : forall e s,
  eval_exprBZ e s = true <-> exprBZ_sem e s.
  induction e.
  simpl.
  split.
  intros.
  eapply Zeq_bool_true.
  auto.
  intros.
  unfold eqb.
  generalize (Zeq_bool_classic (eval_exprZ e s) (eval_exprZ e0 s)); intro X; inversion_clear X.
  rewrite (Zeq_bool_true (eval_exprZ e s) (eval_exprZ e0 s)) in H.
  tauto.
  auto.
  rewrite H.
  apply Zeq_bool_refl.


  simpl.
  split.
  intros.
  generalize (Zgt_cases (eval_exprZ e s) (eval_exprZ e0 s)).
  intros.
  rewrite H in H0.
  omega.
  simpl; intros.
  generalize (Zgt_cases (eval_exprZ e s) (eval_exprZ e0 s)).
  intros.
  assert (Zgt_bool (eval_exprZ e s) (eval_exprZ e0 s)= true \/ Zgt_bool (eval_exprZ e s) (eval_exprZ e0 s)= false).
  elim Zgt_bool.
  left; auto.
  right; auto.
  inversion_clear H1.
  auto.
  rewrite H2 in H0.
  auto.

  simpl.
  split.
  intros.
  generalize (Zge_cases (eval_exprZ e s) (eval_exprZ e0 s)).
  intros.
  rewrite H in H0.
  auto.
  simpl; intros.
  generalize (Zge_cases (eval_exprZ e s) (eval_exprZ e0 s)).
  intros.
  assert (Zge_bool (eval_exprZ e s) (eval_exprZ e0 s)= true \/ Zge_bool (eval_exprZ e s) (eval_exprZ e0 s)= false).
  elim Zge_bool.
  left; auto.
  right; auto.
  inversion_clear H1.
  auto.
  rewrite H2 in H0.
  auto.

  simpl.
  split.
  intros.
  generalize (Zlt_cases (eval_exprZ e s) (eval_exprZ e0 s)).
  intros.
  rewrite H in H0.
  omega.
  simpl; intros.
  generalize (Zlt_cases (eval_exprZ e s) (eval_exprZ e0 s)).
  intros.
  assert (Zlt_bool (eval_exprZ e s) (eval_exprZ e0 s)= true \/ Zlt_bool (eval_exprZ e s) (eval_exprZ e0 s)= false).
  elim Zlt_bool.
  left; auto.
  right; auto.
  inversion_clear H1.
  auto.
  rewrite H2 in H0.
  auto.

  simpl.
  split.
  intros.
  generalize (Zle_cases (eval_exprZ e s) (eval_exprZ e0 s)).
  intros.
  rewrite H in H0.
  auto.
  simpl; intros.
  generalize (Zle_cases (eval_exprZ e s) (eval_exprZ e0 s)).
  intros.
  assert (Zle_bool (eval_exprZ e s) (eval_exprZ e0 s)= true \/ Zle_bool (eval_exprZ e s) (eval_exprZ e0 s)= false).
  elim Zle_bool.
  left; auto.
  right; auto.
  inversion_clear H1.
  auto.
  rewrite H2 in H0.
  auto.

  simpl; split.
  intros.
  generalize (IHe1 s); intro X; inversion_clear X.
  generalize (IHe2 s); intro X; inversion_clear X.
  generalize (orb_prop _ _ H); intro.
  intuition.
  intros.
  apply orb_true_intro.
  generalize (IHe1 s); intro X; inversion_clear X.
  generalize (IHe2 s); intro X; inversion_clear X.
  intuition.

  simpl; split.
  intros.
  generalize (IHe1 s); intro X; inversion_clear X.
  generalize (IHe2 s); intro X; inversion_clear X.
  generalize (andb_prop _ _ H); intro X; inversion_clear X.
  intuition.
  intros.
  apply andb_true_intro.
  generalize (IHe1 s); intro X; inversion_clear X.
  generalize (IHe2 s); intro X; inversion_clear X.
  intuition.

  simpl; split.
  intros.
  assert (eval_exprBZ e s = true \/ eval_exprBZ e s = false).
  elim eval_exprBZ.
  left; auto.
  right; auto.
  inversion_clear H0.
  rewrite H1 in H.
  simpl in H; inversion_clear H.
  elim (classic (exprBZ_sem e s)).
  intros.
  generalize (IHe s); intro X; inversion_clear X.
  intuition.
  rewrite H1 in H5; inversion H5.
  auto.
  intros.
  assert (eval_exprBZ e s = true \/ eval_exprBZ e s = false).
  elim eval_exprBZ.
  left; auto.
  right; auto.
  inversion_clear H0.
  generalize (IHe s); intro X; inversion_clear X.
  intuition.
  rewrite H1; auto.
Qed.

(* A tactic to resolve goals with exprBZ using Coq omega *)


Ltac Simpl_eval := 
  match goal with
    | id: eval_exprZ ?e ?s = ?P |- _ => simpl in id; generalize id; clear id; Simpl_eval
    | id: ?P = eval_exprZ ?e ?s |- _ => simpl in id; generalize id; clear id; Simpl_eval
    | |- _ => (intros || idtac); simpl
  end.

Lemma exprBZ_false_negb_true: forall b s,
  eval_exprBZ b s = false -> eval_exprBZ (neg_bZ b) s = true.
  intros.
  simpl.
  apply negb_false_is_true.
  rewrite negb_elim.
  auto.
Qed.

Lemma exprBZ_true_negb_false: forall b s,
  eval_exprBZ (neg_bZ b) s = true -> eval_exprBZ b s = false.
  intros.
  simpl.
  apply negb_true_is_false.
  simpl in H.
  auto.
Qed.


Ltac
  Eval_b_hyp h :=
  match goal with
    | h: eval_exprBZ ?e ?s = true |- _ => 
      generalize (exprBZ_semantic_good e s); intro eval_b_HypX; inversion_clear eval_b_HypX as [eval_b_HypX1 eval_b_HypX2]; 
        generalize (eval_b_HypX1 h); clear eval_b_HypX1 eval_b_HypX2;
          intro eval_b_HypX; simpl in eval_b_HypX; generalize eval_b_HypX; let x := fresh in intro x; clear eval_b_HypX; clear h
  end
  with
  Eval_b_goal :=
  match goal with
    |  |-  eval_exprBZ ?e ?s = true => 
      generalize (exprBZ_semantic_good e s); intro eval_b_HypX; inversion_clear eval_b_HypX as [eval_b_HypX1 eval_b_HypX2]; 
        apply (eval_b_HypX2); clear eval_b_HypX1 eval_b_HypX2; simpl
    |  |-  eval_exprBZ ?e ?s = false => apply (exprBZ_true_negb_false e s); Eval_b_goal
    | |- _ => idtac
  end
  with
  Eval_b_hyp_clean :=
  match goal with
    | id: eval_exprBZ ?e ?s = true |- _ => Eval_b_hyp id;  Eval_b_hyp_clean
    | id: eval_exprBZ ?e ?s = false |- _ => generalize (exprBZ_false_negb_true e s id); clear id; intro; Eval_b_hyp_clean
    | |- _ => idtac
  end
  with
  Omega_exprb :=
  (intros || idtac); Eval_b_hyp_clean; Eval_b_goal; (tauto || omega || Simpl_eval; OmegaZ).

(* evaluation without environment (None if cannot be evaluated) *)

Fixpoint compute_exprBZ (e:exprBZ): option bool  :=
  match e with 
    eq_bZ e1 e2 => match (compute_exprZ e1) with
                    None => None 
                    | Some b1 => match (compute_exprZ e2) with
                                   None => None
                                   | Some b2 => Some (Zeq_bool b1 b2)
                                 end
                  end
    | gt_bZ e1 e2 => match (compute_exprZ e1) with
                      None => None 
                      | Some b1 => match (compute_exprZ e2) with
                                     None => None
                                     | Some b2 => Some (Zgt_bool b1 b2)
                                   end
                    end
    | ge_bZ e1 e2 => match (compute_exprZ e1) with
                      None => None 
                      | Some b1 => match (compute_exprZ e2) with
                                     None => None
                                     | Some b2 => Some (Zge_bool b1 b2)
                                   end
                    end
    | lt_bZ e1 e2 => match (compute_exprZ e1) with
                      None => None 
                      | Some b1 => match (compute_exprZ e2) with
                                     None => None
                                     | Some b2 => Some (Zlt_bool b1 b2)
                                   end
                    end
    | le_bZ e1 e2 => match (compute_exprZ e1) with
                      None => None 
                      | Some b1 => match (compute_exprZ e2) with
                                     None => None
                                     | Some b2 => Some (Zle_bool b1 b2)
                                   end
                    end
    | or_bZ e1 e2 => match (compute_exprBZ e1) with
                      None => None 
                      | Some b1 => match (compute_exprBZ e2) with
                                     None => None
                                     | Some b2 => Some (orb b1 b2)
                                   end
                    end
    | and_bZ e1 e2 => match (compute_exprBZ e1) with
                       None => None 
                       | Some b1 => match (compute_exprBZ e2) with
                                      None => None
                                      | Some b2 => Some (andb b1 b2)
                                    end
                     end
    | neg_bZ e => match (compute_exprBZ e) with
                   None => None 
                   | Some b => Some (negb b)
                 end
  end.

(* if an boolean expression can be evaluated without environment, this value is correct *)

Lemma compute_exprBZ_correct: forall e z,
  compute_exprBZ e = Some z ->
  (forall s, eval_exprBZ e s = z).
  induction e; intros; simpl in H; try discriminate; simpl.
  generalize (compute_exprZ_correct e0); generalize (compute_exprZ_correct e); intros.
  destruct (compute_exprZ e0); destruct (compute_exprZ e); simpl in H; try discriminate; injection H; intros; subst z; clear H.
  rewrite (H0 z1); auto.
  rewrite (H1 z0); auto.
  generalize (compute_exprZ_correct e0); generalize (compute_exprZ_correct e); intros.
  destruct (compute_exprZ e0); destruct (compute_exprZ e); simpl in H; try discriminate; injection H; intros; subst z; clear H.
  rewrite (H0 z1); auto.
  rewrite (H1 z0); auto.
  generalize (compute_exprZ_correct e0); generalize (compute_exprZ_correct e); intros.
  destruct (compute_exprZ e0); destruct (compute_exprZ e); simpl in H; try discriminate; injection H; intros; subst z; clear H.
  rewrite (H0 z1); auto.
  rewrite (H1 z0); auto.
  generalize (compute_exprZ_correct e0); generalize (compute_exprZ_correct e); intros.
  destruct (compute_exprZ e0); destruct (compute_exprZ e); simpl in H; try discriminate; injection H; intros; subst z; clear H.
  rewrite (H0 z1); auto.
  rewrite (H1 z0); auto.
  generalize (compute_exprZ_correct e0); generalize (compute_exprZ_correct e); intros.
  destruct (compute_exprZ e0); destruct (compute_exprZ e); simpl in H; try discriminate; injection H; intros; subst z; clear H.
  rewrite (H0 z1); auto.
  rewrite (H1 z0); auto.
  destruct (compute_exprBZ e1); destruct (compute_exprBZ e2); simpl in H; try discriminate; injection H; intros; clear H.
  rewrite (IHe1 b); auto.
  rewrite (IHe2 b0); auto.
  destruct (compute_exprBZ e1); destruct (compute_exprBZ e2); simpl in H; try discriminate; injection H; intros; clear H.
  rewrite (IHe1 b); auto.
  rewrite (IHe2 b0); auto.
  destruct (compute_exprBZ e); simpl in H; try discriminate; injection H; intros; clear H.
  rewrite (IHe b); auto.
Qed.


(* factorize a variable into an arithmetic expression *)
Fixpoint exprZ_var_fact' (e: exprZ) (v: nat) {struct e}: (exprZ * exprZ) :=
  match e with
    var_eZ x => if (beq_nat x v) then ((int_eZ 1),(int_eZ 0)) else ((int_eZ 0),(var_eZ x))
    | int_eZ x => ((int_eZ 0),(int_eZ x))
    | e1 +eZ e2 => match (exprZ_var_fact' e1 v) with
                    | (e11,e12) => match (exprZ_var_fact' e2 v) with
                                     | (e21,e22) => ((e11 +eZ e21),(e12 +eZ e22))
                          	   end
                  end
    | e1 -eZ e2 => match (exprZ_var_fact' e1 v) with
                    | (e11,e12) => match (exprZ_var_fact' e2 v) with
                                     | (e21,e22) => ((e11 -eZ e21),(e12 -eZ e22))
                                   end
                  end
    | e1 *eZ e2 => match (exprZ_var_fact' e1 v) with
                    | (e11,e12) => match (exprZ_var_fact' e2 v) with
                                     | (e21,e22) => ((((e12 *eZ e21) +eZ (e11 *eZ e22)) +eZ ((var_eZ v) *eZ (e11 *eZ e21))),(e12 *eZ e22))
                                   end
                  end
  end.

(* factorization is correct (allows to rewrite the term) *)

Lemma exprZ_var_fact'_correct: forall e v e1 e2,
  exprZ_var_fact' e v = (e1,e2) ->
  (forall s, eval_exprZ e s = eval_exprZ (((var_eZ v) *eZ e1) +eZ e2) s).
  induction e; simpl; intros.
  generalize (beq_nat_classic n v); intro X; inversion_clear X.
  rewrite H0 in H.
  injection H; intros; subst e1 e2.
  simpl.
  rewrite (beq_nat_true _ _ H0); omega.
  rewrite H0 in H.
  injection H; intros; subst e1 e2.
  simpl eval_exprZ; omega.
  injection H; intros; subst e1 e2.
  simpl eval_exprZ; omega.
  generalize (IHe1 v); intros; clear IHe1.
  generalize (IHe2 v); intros; clear IHe2.
  destruct (exprZ_var_fact' e1 v); destruct (exprZ_var_fact' e2 v); injection H; intros; subst e0 e3.
  rewrite (H0 e e4); auto; rewrite (H1 e5 e6); auto. 
  simpl; ring.
  generalize (IHe1 v); intros; clear IHe1.
  generalize (IHe2 v); intros; clear IHe2.
  destruct (exprZ_var_fact' e1 v); destruct (exprZ_var_fact' e2 v); injection H; intros; subst e0 e3.
  rewrite (H0 e e4); auto; rewrite (H1 e5 e6); auto. 
  simpl; ring.
  generalize (IHe1 v); intros; clear IHe1.
  generalize (IHe2 v); intros; clear IHe2.
  destruct (exprZ_var_fact' e1 v); destruct (exprZ_var_fact' e2 v); injection H; intros; subst e0 e3.
  rewrite (H0 e e4); auto; rewrite (H1 e5 e6); auto. 
  simpl; ring.
Qed.


(* arithmetic constraints are all stored in a normal formed: an arithmetic expression e

  Its semantic is: e <= 0
*)

Definition constraint := exprZ.

(* TODO: pour les gens oisifs refaire en := c >>=Z (int_eZ 0) *)
Definition constraint_semantic (c: constraint) : exprBZ := (int_eZ 0) >>=Z c.

(* Simplify a constraint for a given list a variables 

  Simplify by means that it tries to replace coefficient of variables by a value (evaluation without environment)

*)

Fixpoint simpl_varlist_constraint (c: constraint) (v: list nat) {struct v}: constraint :=
  match v with
    nil => match (compute_exprZ c) with
             None => c
             | Some val => int_eZ val
           end
    | hd::tl => match (exprZ_var_fact' c hd) with
                  (e1,e2) => match (compute_exprZ e1) with
                               None => (e1 *eZ (var_eZ hd)) +eZ (simpl_varlist_constraint e2 tl)
                               | Some val => if (Zeq_bool val 0%Z) then 
                                 (simpl_varlist_constraint e2 tl)
                                 else
                                   ((int_eZ val) *eZ (var_eZ hd)) +eZ (simpl_varlist_constraint e2 tl)
                             end
                end
  end.

(* Some lemmas to make cleaner destruct in proofs *)

Lemma simpl_varlist_constraint_destruct: forall c v,
  exists c', simpl_varlist_constraint c v = c'.
  intros.
  exists (simpl_varlist_constraint c v).
  auto.
Qed.

Lemma compute_exprZ_destruct: forall c,
  ((exists z, compute_exprZ c = Some z) \/ compute_exprZ c = None).
  intros.
  destruct compute_exprZ.
  left; exists z; auto.
  right; auto.
Qed.

Lemma exprZ_num_var_fact'_destruct: forall e v,
  exists e1, exists e2, exprZ_var_fact' e v = (e1,e2).
  intros.
  destruct exprZ_var_fact'.
  exists e0; exists e1; auto.
Qed. 

(* an arithmetic expression and its simplification evaluates similarly for every environment *)

Lemma simpl_varlist_constraint_correct: forall v c c',
  simpl_varlist_constraint c v = c' ->
  (forall s, eval_exprZ c s = eval_exprZ c' s).
  induction v.
  simpl.
  intros.
  generalize (compute_exprZ_destruct c); intro X; inversion_clear X.
  inversion_clear H0.
  rewrite H1 in H.
  rewrite <- H.
  rewrite (compute_exprZ_correct _ _ H1).
  auto.
  rewrite H0 in H.
  subst c; auto.
  simpl in IHv.
  simpl; intros.
  generalize (exprZ_num_var_fact'_destruct c a); intro X; inversion_clear X.
  inversion_clear H0.
  rewrite H1 in H.
  generalize (exprZ_var_fact'_correct _ _ _ _ H1 s); clear H1; intros.
  generalize (compute_exprZ_destruct x); intro X; inversion_clear X.
  inversion_clear H1.
  rewrite H2 in H.
  generalize (compute_exprZ_correct _ _ H2 s); intros.
  generalize (Zeq_bool_classic x1 0); intro X; inversion_clear X.
  rewrite H3 in H.
  generalize (IHv _ _ H s); intros.
  simpl in H0.
  generalize (Zeq_bool_true _ _ H3); intros.
  rewrite H5 in H1; rewrite H1 in H0.
  rewrite H0.
  cutrewrite (nth a s 0 * 0 + eval_exprZ x0 s = eval_exprZ x0 s)%Z.
  auto.
  intuition.
  rewrite H3 in H.
  rewrite H0; rewrite <- H.
  simpl.
  rewrite <- H1.
  generalize (simpl_varlist_constraint_destruct x0 v); intros. 
  inversion_clear H4.
  generalize (IHv _ _ H5 s); intros.
  rewrite H5.
  intuition.
  rewrite H1 in H.
  subst c'.
  simpl.
  rewrite H0.
  simpl.
  generalize (simpl_varlist_constraint_destruct x0 v); intros. 
  inversion_clear H.
  generalize (IHv _ _ H2 s); intros.
  intuition.
Qed.

(* Simplification of a constraints: simplification for all its variables *)

Definition simpl_constraint (c: constraint) : constraint :=
  simpl_varlist_constraint c (exprZ_var c).

(* simplification preserves evaluation *)

Lemma simpl_constraint_correct: forall c,
  (forall s, eval_exprZ c s = eval_exprZ (simpl_constraint c) s).
  intros.
  unfold simpl_constraint.
  generalize (simpl_varlist_constraint_destruct c (exprZ_var c)); intros.
  inversion_clear H.
  rewrite H0.
  eapply simpl_varlist_constraint_correct.
  apply H0.
Qed.

(* simplification of a constraint for a variables with simplification of its coefficient *)

Definition simpl_constraint_var' (c: constraint) (v: nat): constraint :=
  match (exprZ_var_fact' c v) with
    (e1,e2) => match (compute_exprZ (simpl_constraint e1)) with
                 None => (e1 *eZ (var_eZ v)) +eZ e2
                 | Some val => if (Zeq_bool val 0%Z) then 
                   e2
                   else
                     ((int_eZ val) *eZ (var_eZ v)) +eZ e2
               end
  end.

(* This simplification preserves evaluation for every environments *)

Lemma simpl_constraint_var'_correct: forall c v,
  (forall s, eval_exprZ c s = eval_exprZ (simpl_constraint_var' c v) s).
  intros.
  unfold simpl_constraint_var'.
  generalize (exprZ_var_fact'_correct c v); intros.
  destruct (exprZ_var_fact' c v).
  generalize (compute_exprZ_correct (simpl_constraint e)); intros.
  destruct (compute_exprZ (simpl_constraint e)).
  generalize (Zeq_bool_classic z 0); intro X; inversion_clear X.
  generalize (Zeq_bool_true _ _ H1); intros; subst z.
  rewrite H1.
  lapply (H e e0); intros; auto.
  rewrite H2; clear H2.
  lapply (H0 0%Z); intros; auto.
  simpl.
  rewrite (simpl_constraint_correct e).
  rewrite H2.
  ring.
  rewrite H1.
  lapply (H e e0); intros; auto.
  rewrite H2; clear H2.
  simpl.
  rewrite (simpl_constraint_correct e).
  lapply (H0 z); intros; auto.
  rewrite H2; ring.
  lapply (H e e0); intros; auto.
  rewrite H1; simpl; ring.
Qed.

(* Previous simplification for a set of variables *)

Fixpoint simpl_constraint_var_list' (c: constraint) (v: list nat) {struct v}: constraint :=
  match v with
    nil => simpl_constraint c
    | hd::tl => simpl_constraint_var_list' (simpl_constraint_var' c hd) tl
  end.

(* Previous simplification for a set of variables preserves evaluation *)

Lemma simpl_constraint_var_list'_correct: forall v c,
  (forall s, eval_exprZ c s = eval_exprZ (simpl_constraint_var_list' c v) s).
  induction v; simpl.
  intros; rewrite simpl_constraint_correct; auto.
  intros; rewrite <- IHv.
  rewrite (simpl_constraint_var'_correct c a); auto.
Qed.

(* Simplification for all the variables of the constraints *)

Definition simpl_constraint' (c: constraint) : constraint :=
  simpl_constraint_var_list' c (exprZ_var c).

(* This simplification preserves invariants *)

Lemma simpl_constraint'_correct: forall c,
  (forall s, eval_exprZ c s = eval_exprZ (simpl_constraint' c) s).
  intros; unfold simpl_constraint'; rewrite <- simpl_constraint_var_list'_correct; auto.
Qed.

(* Factorization with simplification of coefficients *)

Definition exprZ_var_fact (e: exprZ) (v: nat): (exprZ * exprZ) :=
  match (exprZ_var_fact' e v) with
    (e1,e2) => (simpl_constraint e1,simpl_constraint e2)
  end.

(* From factorization we can build a constraints with the same evaluation value *)

Lemma exprZ_var_fact_correct: forall c v e1 e2,
  exprZ_var_fact c v = (e1,e2) ->
  (forall s, eval_exprZ c s = eval_exprZ (((var_eZ v) *eZ e1) +eZ e2) s).
  intros.
  unfold exprZ_var_fact in H.
  generalize (exprZ_num_var_fact'_destruct c v); intro X.
  inversion_clear X.
  inversion_clear H0.
  rewrite H1 in H.
  generalize (exprZ_var_fact'_correct _ _ _ _ H1 s); intro.
  injection H; intros.
  generalize (simpl_constraint_correct x0 s); intros.
  generalize (simpl_constraint_correct x s); intros.
  rewrite H0.
  simpl.
  rewrite <- H2; rewrite <- H3.
  rewrite <- H4; rewrite <- H5.
  auto.
Qed.

(* Variable elimination:
  c1 and c2 are two constraints containing the variable v.
  By calculating the coefficients of v we can compute if the variables can be eliminated
  (for details on the elimination mechanism see the following lemma)
*)

Definition elim_var_constraint (c1 c2: constraint) (v: nat) : constraint :=
  match (exprZ_var_fact c1 v) with
    (e11,e12) => match (exprZ_var_fact c2 v) with
                   (e21,e22) => match (compute_exprZ (simpl_constraint e11)) with
                                  None => c2
                                  | Some e11' => match (compute_exprZ (simpl_constraint e21)) with
                                                   None => c2
                                                   | Some e21' => if (andb (Zlt_bool e11' 0%Z) (Zlt_bool 0%Z e21')) then
                                                     (((e21 *eZ e12) -eZ (e11 *eZ e22)))
                                                     else (if (andb (Zlt_bool 0%Z e11') (Zlt_bool e21' 0%Z)) then
                                                       (((e11 *eZ e22) -eZ (e21 *eZ e12))) else c2)
                                                 end   
                                end   
                 end
  end.


(* The principles of elimination of var x
       i) isolate the term in x and the rest
       ii) make the coefficient of x equal (must know the sign of x coefficients for the multiplication)
       iii) use transitivity of <=

  *)

Lemma presburger_lemma1: forall a1 b1 a2 b2 x,
  (a1 < 0 )%Z ->
  (0 < a2 )%Z ->
  (0 >= x * a1 + b1)%Z ->
  (0 >= x * a2 + b2)%Z ->
  (a1 * b2 >= a2 * b1)%Z.
  intros.
  assert (b1 <= - (x * a1))%Z.
  intuition.
  assert ((x * a2) <= - b2)%Z.
  intuition.
  assert (b1 * a2 <= - (x * a1) * a2)%Z.
  intuition.
  assert ((x * a2) * -a1 <= - b2 * - a1)%Z.
  intuition.
  assert (x * a2 * - a1 = -(x * a1) * a2)%Z.
  ring.
  rewrite H7 in H6; clear H7.
  assert (b1 * a2 <= - b2 * - a1)%Z.
  intuition.
  assert (-b2 * - a1 = a1 * b2)%Z.
  ring.
  rewrite H8 in H7; clear H8.
  cutrewrite (a2 * b1 = b1 * a2)%Z.
  intuition.
  intuition.
Qed.

(* The two constraints allowing the variables elimination implies the constraint resulting from elimination  *)
Lemma elim_var_constraint_correct: forall c1 c2 v,
  (forall s, (s |bZ= ((constraint_semantic c1) &&&Z (constraint_semantic c2))) -> (s |bZ= (constraint_semantic (elim_var_constraint c1 c2 v)))).
  intros.
  unfold elim_var_constraint.
  generalize (exprZ_var_fact_correct c1 v); intros.
  generalize (exprZ_var_fact_correct c2 v); intros.
  destruct (exprZ_var_fact c1 v); destruct (exprZ_var_fact c2 v).
  generalize (compute_exprZ_correct (simpl_constraint e)); intros.
  generalize (compute_exprZ_correct (simpl_constraint e1)); intros.
  destruct (compute_exprZ (simpl_constraint e)); destruct (compute_exprZ (simpl_constraint e1)); auto.
  generalize (Zlt_cases z 0); intros.
  generalize (Zlt_cases 0 z0); intros.
  destruct (Zlt_bool z 0); destruct (Zlt_bool 0 z0).
  simpl constraint_semantic.
  assert ((s |bZ= constraint_semantic c1) /\ (s |bZ= constraint_semantic c2)).
  split; Omega_exprb.
  inversion_clear H6.
  simpl in H7.
  simpl in H8.
  rewrite (H0 e e0) in H7; auto.
  rewrite (H1 e1 e2) in H8; auto.
  clear H0 H1.
  simpl in H7; simpl in H8.
  simpl.
  rewrite (simpl_constraint_correct e) in H7; rewrite (H2 z) in H7; auto.
  rewrite (simpl_constraint_correct e1) in H8; rewrite (H3 z0) in H8; auto.
  rewrite (simpl_constraint_correct e); rewrite (H2 z); auto.
  rewrite (simpl_constraint_correct e1); rewrite (H3 z0); auto.
  generalize (Zge_cases 0%Z ((nth v s 0%Z) * z + eval_exprZ e0 s)); intros.
  rewrite H7 in H0; clear H7.
  generalize (Zge_cases 0 (nth v s 0%Z * z0 + eval_exprZ e2 s)); intros.
  rewrite H8 in H1; clear H8.
  generalize (presburger_lemma1 _ _ _ _ _ H4 H5 H0 H1); intros.
  generalize (Zge_cases 0 (z0 * eval_exprZ e0 s - z * eval_exprZ e2 s)); intros.
  destruct (Zge_bool 0 (z0 * eval_exprZ e0 s - z * eval_exprZ e2 s)).
  auto.
  intuition.
  simpl constraint_semantic.
  generalize (Zlt_cases 0 z); intros.
  destruct (Zlt_bool 0 z).
  intuition.
  simpl constraint_semantic.
  Omega_exprb.
  simpl constraint_semantic.
  generalize (Zlt_cases 0 z); intros.
  destruct (Zlt_bool 0 z).
  generalize (Zlt_cases z0 0); intros.
  destruct (Zlt_bool z0 0).
  intuition.
  simpl constraint_semantic.
  Omega_exprb.
  simpl constraint_semantic.
  Omega_exprb.
  simpl constraint_semantic.
  generalize (Zlt_cases 0 z); intros.
  destruct (Zlt_bool 0 z).
  generalize (Zlt_cases z0 0); intros.
  destruct (Zlt_bool z0 0).
  assert ((s |bZ= constraint_semantic c1) /\ (s |bZ= constraint_semantic c2)).
  split; Omega_exprb.
  simpl constraint_semantic.
  clear H4 H5.
  inversion_clear H8.
  simpl in H4.
  simpl in H5.
  rewrite (H0 e e0) in H4; auto.
  rewrite (H1 e1 e2) in H5; auto.
  simpl in H4; simpl in H5.
  rewrite (simpl_constraint_correct e) in H4; rewrite (H2 z) in H4; auto.
  rewrite (simpl_constraint_correct e1) in H5; rewrite (H3 z0) in H5; auto.
  simpl.
  rewrite (simpl_constraint_correct e); rewrite (H2 z); auto.
  rewrite (simpl_constraint_correct e1); rewrite (H3 z0); auto.
  generalize (Zge_cases 0 (nth v s 0%Z * z + eval_exprZ e0 s)); intros.
  rewrite H4 in H8; clear H4.
  generalize (Zge_cases 0 (nth v s 0%Z * z0 + eval_exprZ e2 s)); intros.
  rewrite H5 in H4; clear H5.
  generalize (presburger_lemma1 _ _ _ _ _ H7 H6 H4 H8); intros.
  generalize (Zge_cases 0 (z * eval_exprZ e2 s - z0 * eval_exprZ e0 s)); intros.
  destruct (Zge_bool 0 (z * eval_exprZ e2 s - z0 * eval_exprZ e0 s)).
  auto.
  intuition.
  simpl constraint_semantic.
  Omega_exprb.
  simpl constraint_semantic.
  Omega_exprb.
  Omega_exprb.
  Omega_exprb.
  Omega_exprb.
Qed.

(* Some simple lemmas for proof readability *)

Lemma exprBZ_neg_involutive: forall a,
  (forall s, eval_exprBZ (!Z (!Z a)) s = eval_exprBZ a s).
  intros.
  simpl.
  rewrite negb_elim.
  auto.
Qed.


Lemma evalB_destruct: forall b s,
  eval_exprBZ b s = true \/ eval_exprBZ b s = false.
  intros.
  destruct eval_exprBZ; auto.
Qed.

Lemma exprBZ_eq: forall b1 b2 s,
  ((s |bZ= b1) <-> (s |bZ= b2)) -> (eval_exprBZ b1 s = eval_exprBZ b2 s).
  intros.
  inversion_clear H.
  generalize (evalB_destruct b1 s); intro X; inversion_clear X.
  generalize (evalB_destruct b2 s); intro X; inversion_clear X.
  rewrite H; rewrite H2; auto.
  intuition.
  rewrite H3 in H2; discriminate.
  generalize (evalB_destruct b2 s); intro X; inversion_clear X.
  intuition.
  rewrite H3 in H; discriminate.
  rewrite H; rewrite H2; auto.
Qed.

Lemma exprBZ_eq': forall b1 b2 s,
  (eval_exprBZ b1 s = eval_exprBZ b2 s) -> ((s |bZ= b1) <-> (s |bZ= b2)).
  intros.
  split; intros.
  rewrite H in H0.
  auto.
  rewrite <- H in H0.
  auto.
Qed.


(* Transformation to form normal 
  
  Captures by dnf_expr_rec (this function stop in a sub-term if a simplification is possible)

  Follow a hierarchical simplification
  i) then [faire rentrer les negations]
  ii) on top distributivity of /\ for \/
  iii) eventually put all the term to a form 
                  a >= b
*)

Definition dnf_expr_b1 (a:exprBZ) : exprBZ :=
  match a with
    | e1 ==Z e2 => (e2 >>=Z e1) &&&Z (e1 >>=Z e2)
    | e1 >>Z e2 =>  e1 >>=Z  (e2 +eZ (int_eZ 1))
    | e1 <<Z e2 =>  e2 >>=Z  (e1 +eZ (int_eZ 1))
    | e1 <<=Z e2 =>  e2 >>=Z  e1
    | _ => a
  end.

Definition dnf_expr_b2 (a:exprBZ) : exprBZ :=
  match a with
    |  a1 &&&Z (a2 |||Z a3) => (a1 &&&Z a2) |||Z (a1 &&&Z a3)
    | (a2 |||Z a3) &&&Z a1 => (a1 &&&Z  a2) |||Z  (a1  &&&Z  a3)
    | _ => dnf_expr_b1 a
  end.

Lemma dnf_expr_b2_correct: forall a,
  (forall s, eval_exprBZ a s = eval_exprBZ (dnf_expr_b2 a) s).
  destruct a; (intros; apply exprBZ_eq; split; (intros; try Omega_exprb)).
  destruct a1; destruct a2; try Omega_exprb.
  destruct a1; destruct a2; try Omega_exprb.
Qed.

Definition dnf_expr_b3 (a:exprBZ) : exprBZ :=
  match a with
    | !Z (!Z a') => (a')
    | !Z (a1 &&&Z a2) => (!Za1) |||Z (!Za2)
    | !Z (a1 |||Z a2) => (!Z a1) &&&Z (!Z a2)
    | !Z (e1 >>Z e2) => (e2 >>=Z e1)
    | !Z (e1 >>=Z e2) => (e2 >>Z e1)
    | !Z (e1 <<Z e2) => (e1 >>=Z e2)
    | !Z (e1 <<=Z e2) => (e1 >>Z e2)
    | !Z (e1 ==Z e2) => (e2 >>Z e1) |||Z (e1 >>Z e2)
    | _ => dnf_expr_b2 a
  end.

Lemma dnf_expr_b3_correct: forall a,
  (forall s, eval_exprBZ a s = eval_exprBZ (dnf_expr_b3 a) s).
  destruct a; (intros; apply exprBZ_eq; split; (intros; try Omega_exprb)).
  destruct a1; destruct a2; try Omega_exprb.
  destruct a1; destruct a2; try Omega_exprb.
  destruct a; try Omega_exprb.
  simpl.
  simpl in H.
  destruct (eval_exprBZ a1 s); destruct (eval_exprBZ a2 s); (auto || (simpl in H; inversion H)).
  rewrite exprBZ_neg_involutive in H; auto.
  destruct a; try Omega_exprb.
Qed.


Fixpoint dnf_expr_rec (a: exprBZ) : exprBZ :=
  match a with
    | a1 &&&Z a2 => dnf_expr_b2 ((dnf_expr_rec a1) &&&Z (dnf_expr_rec a2))
    | a1 |||Z a2 => (dnf_expr_rec a1) |||Z (dnf_expr_rec a2)
    | _ =>  dnf_expr_b3 a
  end.


(* The normal form preserves evaluation *)

Lemma dnf_expr_rec_correct: forall a,
  (forall s, eval_exprBZ a s = eval_exprBZ (dnf_expr_rec a) s).
  induction a; (intros; apply exprBZ_eq; split; (intros; try Omega_exprb)).
  (* |||Z *)
  cut (s |bZ= ((dnf_expr_rec a1) |||Z (dnf_expr_rec a2))).
  intuition.
  simpl in H.
  simpl eval_exprBZ.
  rewrite <- IHa1; rewrite <- IHa2; auto.
  assert (s |bZ= (((dnf_expr_rec a1) |||Z (dnf_expr_rec a2)))).
  intuition.
  simpl eval_exprBZ in H0.
  simpl.
  rewrite IHa1; rewrite IHa2; auto.
  (* &&&Z *)
  cut (s |bZ= (dnf_expr_b2 ((dnf_expr_rec a1) &&&Z (dnf_expr_rec a2)))).
  intuition.
  rewrite <- dnf_expr_b2_correct.
  simpl in H.
  simpl eval_exprBZ.
  rewrite <- IHa1; rewrite <- IHa2; auto.
  assert (s |bZ= (dnf_expr_b2 ((dnf_expr_rec a1) &&&Z (dnf_expr_rec a2)))).
  intuition.
  rewrite <- dnf_expr_b2_correct in H0.
  simpl eval_exprBZ in H0.
  simpl.
  rewrite IHa1; rewrite IHa2; auto.
  (* !Z *)
  rewrite dnf_expr_b3_correct in H.
  intuition.
  rewrite dnf_expr_b3_correct.
  intuition.
Qed.

(* However as a normalization stop the function for a sub-term, we must call it n times.
  n is compute by the following function
*)

Fixpoint expr_b_size (a: exprBZ) : nat :=
  match a with 
    | eq_bZ e1 e2 => 1
    | ge_bZ e1 e2 => 1
    | gt_bZ e1 e2 => 1
    | le_bZ e1 e2 => 2
    | lt_bZ e1 e2 => 2
    | and_bZ e1 e2 => 1 + ((expr_b_size e1) * (expr_b_size e2))
    | or_bZ e1 e2 => 1 + ((expr_b_size e1) + (expr_b_size e2))
    | neg_bZ e => 1 + (expr_b_size e)
  end.

(* we call the normalization function size times *)

Fixpoint dnf_expr_fp (a: exprBZ) (size: nat) {struct size}: exprBZ :=
  match size with
    0 => dnf_expr_rec a
    | S n => dnf_expr_fp (dnf_expr_rec a) n
  end.

(* This recursive normalization preserves evaluation *)

Lemma dnf_expr_fp_correct: forall n a,
  (forall s, eval_exprBZ a s = eval_exprBZ (dnf_expr_fp a n) s).
  induction n.
  simpl; intros.
  apply dnf_expr_rec_correct.
  simpl; intros.
  rewrite <- (IHn (dnf_expr_rec a) s).
  apply dnf_expr_rec_correct.
Qed.

(* Normalization of a boolean expression  *)

Definition dnf_expr_b (a:exprBZ) : exprBZ := dnf_expr_fp a (expr_b_size a).

(* Normalization preserves evaluation *)

Lemma dnf_expr_b_correct: forall a,
  (forall s, eval_exprBZ a s = eval_exprBZ (dnf_expr_b a) s).
  intros.
  unfold dnf_expr_b.
  apply dnf_expr_fp_correct.
Qed.

(* Translation for arithmetic constraint to normalized constraint (None when no possible translation) *)

Definition dnf_expr_b2constraint (b: exprBZ) : option constraint :=
  match b with
    e1 >>=Z e2 => Some (e2 -eZ e1)
    | _ => None
  end.

(* If possible, Semantic Translation evaluate similarly as the original expression *)

Lemma dnf_expr_b2constraint_correct: forall b b',
  dnf_expr_b2constraint b = Some b' ->
  (forall s, eval_exprBZ b s = eval_exprBZ (constraint_semantic b') s).
  destruct b; (simpl; intros; try inversion H).
  simpl.
  generalize (Zge_cases (eval_exprZ e s) (eval_exprZ e0 s)); intros.
  generalize (Zge_cases 0 (eval_exprZ e0 s - eval_exprZ e s)); intros.

  destruct (Zge_bool (eval_exprZ e s) (eval_exprZ e0 s)); destruct (Zge_bool 0 (eval_exprZ e0 s - eval_exprZ e s)); auto.
  assert (0 >= eval_exprZ e0 s - eval_exprZ e s)%Z.
  intuition.
  contradiction.
  assert (0 < eval_exprZ e0 s - eval_exprZ e s)%Z.
  intuition.
  contradiction.
Qed.

(* boolean evaluation of a constraint without environment  *)

Definition eval_constraint (c : constraint) : option bool :=
   match compute_exprZ c with
     | Some z => Some (Zge_bool 0 z)
     | _ => None
   end.

(* if possible, evaluation is valid for every environment  *)

Lemma eval_constraint2constraint_semantic: forall c b,
  eval_constraint c = Some b ->
  (forall s, eval_exprBZ (constraint_semantic c) s = b).
intros.
generalize (compute_exprZ_correct c); intros.
unfold eval_constraint in H.
destruct (compute_exprZ c); try discriminate.
simpl.
injection H; intros; subst b.
rewrite (H0 z); auto.
Qed.

(**********************************)

(* And list are abstraction for conjunction of constraints *)

Definition andlist  := list constraint.

Definition true_b : exprBZ := int_eZ 0 ==Z int_eZ 0.
Definition false_b : exprBZ := int_eZ 0 ==Z int_eZ 1.

(* And list semantic *)

Fixpoint andlist_semantic (l: andlist) : exprBZ:=
  match l with
    nil => true_b
    | hd::tl => (constraint_semantic hd) &&&Z (andlist_semantic tl)
  end.

(* Translation of an boolean expression to an andlist *)

Fixpoint dnf_expr_b2andlist (b: exprBZ) : option andlist :=
  match b with
    b1 &&&Z b2 => option_app (dnf_expr_b2andlist b1) (dnf_expr_b2andlist b2)
    | _ => match (dnf_expr_b2constraint b) with
             None => None
             | Some b' => Some (b'::nil)
           end
  end.

(* Some tools lemma for andlist *)

Lemma dnf_expr_b2andlist_notempty: forall b b',
  dnf_expr_b2andlist b = Some b' -> 
  b' <> nil.
  induction b; intros; simpl in H; try discriminate.
  injection H; intros; subst b'; clear H.
  red; intros; discriminate.
  destruct (dnf_expr_b2andlist b1); destruct (dnf_expr_b2andlist b2); simpl in H; try discriminate.
  injection H; intros; subst b'; clear H.
  generalize (IHb1 a); intros.
  intuition.
  destruct a; intuition; discriminate.
Qed.

Lemma andlist_semantic_app: forall l1 l2 s,
  eval_exprBZ (andlist_semantic (l1 ++ l2)) s =
  eval_exprBZ ((andlist_semantic l1) &&&Z (andlist_semantic l2)) s.
  induction l1.
  intuition.
  intros.
  simpl andlist_semantic.
  apply exprBZ_eq; split; intros; try Omega_exprb.
  assert (A1: (s |bZ= constraint_semantic a) /\ s |bZ= andlist_semantic (l1 ++ l2)).
  split; Omega_exprb.
  inversion_clear A1.
  rewrite IHl1 in H1.
  Omega_exprb.
  cut ((s |bZ= constraint_semantic a) /\ s |bZ= andlist_semantic (l1 ++ l2)).
  intros.
  inversion_clear H0.
  Omega_exprb.
  split; try Omega_exprb.
  rewrite IHl1.
  Omega_exprb.
Qed.

(* If a translation exists, its semantic preserves the evaluation of the original term *)

Lemma dnf_expr_bool2andlist_correct: forall b b',
  dnf_expr_b2andlist b = Some b' ->
  (forall s, eval_exprBZ b s = eval_exprBZ (andlist_semantic b') s).
  induction b; intros; simpl in H; try discriminate.
  injection H; intros; subst b'.
  simpl andlist_semantic.
  cutrewrite (eval_exprBZ (constraint_semantic (e0 -eZ e) &&&Z true_b) s = eval_exprBZ (constraint_semantic (e0 -eZ e)) s).
  rewrite (dnf_expr_b2constraint_correct (e >>=Z e0) (e0 -eZ e)).
  auto.
  auto.
  apply exprBZ_eq; split; intros; Omega_exprb.
  destruct (dnf_expr_b2andlist b1); destruct (dnf_expr_b2andlist b2); simpl in H; try discriminate.
  injection H; intros; subst b'; clear H.
  rewrite andlist_semantic_app.
  simpl.
  rewrite (IHb2  a0); auto.
  rewrite (IHb1 a); auto.
Qed.

(* Given a constraint c, a list a constraint l and a variable v,
  we build the list l' of constraints such that it contains
  i) all the constraint of l where v does not appear
  ii)  all the constraint of l where v appears but cannot be eliminated using c
  iii) all constraint of l were v appear and have been eliminated using c
*)

Fixpoint elim_var_andlist' (c: constraint) (l l': andlist) (v: nat) {struct l} : andlist := 
  match l with
     nil => l'
     | hd::tl => if (inb beq_nat (exprZ_var (simpl_constraint_var_list' hd (v::nil))) v) then
                            (elim_var_andlist' c tl ((elim_var_constraint c hd v)::l') v)
                      else
                            (elim_var_andlist' c tl l' v)
  end.

(*   *)

Lemma elim_var_andlist'_correct: forall l l' c v,
  (forall s, (s |bZ=  (andlist_semantic (c::l ++ l'))) -> 
                s |bZ=  (andlist_semantic (elim_var_andlist' c l l' v))).
induction l.
simpl andlist_semantic.
intros; Omega_exprb.
simpl andlist_semantic; simpl andlist_semantic in IHl.
intros.
destruct inb.
assert (
(s |bZ= constraint_semantic c) /\
(s |bZ= constraint_semantic a) /\
(s |bZ= andlist_semantic (l++l'))
).
split; try Omega_exprb.
split; try Omega_exprb.
inversion_clear H0.
inversion_clear H2.
rewrite andlist_semantic_app in H3.
apply IHl.
cut(
(s |bZ= constraint_semantic c) /\
(s |bZ=  andlist_semantic (l ++ elim_var_constraint c a v :: l'))
).
intros.
inversion_clear H2.
Omega_exprb.
split; try Omega_exprb.
rewrite andlist_semantic_app.
cut(
(s |bZ=  andlist_semantic l) /\
(s |bZ=  andlist_semantic (elim_var_constraint c a v :: l'))
).
intro X; inversion_clear X.
Omega_exprb.
split; try Omega_exprb.
simpl andlist_semantic.
cut(
(s |bZ= constraint_semantic (elim_var_constraint c a v)) /\
(s |bZ= andlist_semantic l')
).
intro X; inversion_clear X.
Omega_exprb.
split; try Omega_exprb.
apply elim_var_constraint_correct.
Omega_exprb.
apply IHl.
Omega_exprb.
Qed.

(* For a given variable v, find the next constraint containing v, and
use it to eliminate v on the rest of the list *)

Fixpoint elim_var_andlist (l l': andlist) (v: nat) {struct l} : andlist :=
  match l with
     nil => l'
     | hd::tl => if (inb beq_nat (exprZ_var (simpl_constraint_var_list' hd (v::nil))) v) then
                          (elim_var_andlist tl (l' ++ (elim_var_andlist' hd tl nil v)) v)
                      else
                          (elim_var_andlist tl (hd::l') v)
  end.

(*  *)

Lemma elim_var_andlist_correct: forall l l' v,
  (forall s, (s |bZ=  (andlist_semantic (l ++ l'))) -> 
                s |bZ=  (andlist_semantic (elim_var_andlist l l' v))).
induction l.
simpl andlist_semantic.
intuition.
simpl andlist_semantic.
intros.
destruct inb.
apply IHl.
rewrite (andlist_semantic_app).
cut(
(s |bZ= andlist_semantic l) /\
(s |bZ= andlist_semantic (l' ++ elim_var_andlist' a l nil v))
).
intros.
inversion_clear H0.
Omega_exprb.
assert(
(s |bZ= constraint_semantic a) /\
(s |bZ= andlist_semantic (l ++ l'))
).
split; try Omega_exprb.
split; try Omega_exprb.
inversion_clear H0.
rewrite (andlist_semantic_app) in H2.
Omega_exprb.
inversion_clear H0.
rewrite (andlist_semantic_app).
rewrite (andlist_semantic_app) in H2.
cut (
(s |bZ= andlist_semantic l') /\
(s |bZ= andlist_semantic (elim_var_andlist' a l nil v))
).
intro X; inversion_clear X.
Omega_exprb.
split; try Omega_exprb.
apply elim_var_andlist'_correct.
simpl andlist_semantic.
rewrite <- app_nil_end.
Omega_exprb.
apply IHl.
rewrite (andlist_semantic_app).
simpl andlist_semantic.
assert (
(s |bZ= constraint_semantic a) /\
(s |bZ=  andlist_semantic (l ++ l'))
).
split; try Omega_exprb.
inversion_clear H0.
rewrite (andlist_semantic_app) in H2.
Omega_exprb.
Qed.

(* simplification of every constraints of an andlist  *)

Fixpoint simpl_andlist (l: andlist): andlist :=
   match l with
      nil => nil
      | hd::tl => (simpl_constraint' hd )::(simpl_andlist tl )
   end.

(* simplification preserves evaluation *)

Lemma simpl_andlist_correct: forall l ,
 (forall s, eval_exprBZ (andlist_semantic l) s =   eval_exprBZ (andlist_semantic (simpl_andlist l)) s).
induction l.
auto.
simpl andlist_semantic.
intros.
apply exprBZ_eq; split; intros.
assert (
(s |bZ= constraint_semantic a) /\
(s |bZ= andlist_semantic l)
).
split; Omega_exprb.
clear H; inversion_clear H0.
cut(
(s |bZ= constraint_semantic (simpl_constraint' a)) /\
(s |bZ= andlist_semantic (simpl_andlist l))
).
intros.
inversion_clear H0.
Omega_exprb.
split.
simpl.
rewrite <- simpl_constraint'_correct.
simpl in H; auto.
rewrite <- IHl.
auto.
assert(
(s |bZ= constraint_semantic (simpl_constraint' a)) /\
(s |bZ= andlist_semantic (simpl_andlist l))
).
split; Omega_exprb.
clear H; inversion_clear H0.
rewrite <- IHl in H1.
assert (s |bZ= constraint_semantic a).
simpl in H.
rewrite <- simpl_constraint'_correct in H.
auto.
Omega_exprb.
Qed.

(* evaluation of a not empty andlist without environment *)


Fixpoint eval_andlist' (a: andlist) : option bool :=
   match a with
     nil => Some true
     | hd::tl => match (eval_constraint hd) with
                         None => None
                         | Some b => andb_option (Some b) (eval_andlist' tl)
                      end
   end.

(* if evaluation was possible, it holds for every environment *)

Lemma eval_andlist'2andlist_semantic: forall a b,
  eval_andlist' a = Some b ->
  (forall s, eval_exprBZ (andlist_semantic a) s = b).
induction a.
simpl.
intros.
injection H; auto.
simpl eval_andlist'.
intros.
generalize (eval_constraint2constraint_semantic a); intros.
destruct (eval_constraint a).
simpl in H0.
simpl.
rewrite (H0 b0); auto.
destruct (eval_andlist' a0).
lapply  (IHa b1); intros; auto.
rewrite H1.
injection H; auto.
inversion H.
inversion H.
Qed.

(* Evaluation of an andlist without environment. If empty => error !Z*)

Definition eval_andlist (a: andlist) : option bool :=
  if (beq_nat (length a) 0) then
      None 
  else     
      eval_andlist' a.

(* if the evaluation is possible, it holds for every environment *)

Lemma eval_andlist2andlist_semantic: forall a b,
  eval_andlist a = Some b ->
  (forall s, eval_exprBZ (andlist_semantic a) s = b).
intros.
unfold eval_andlist in H.
generalize (beq_nat_classic (length a) 0); intro X; inversion_clear X.
rewrite H0 in H; discriminate.
rewrite H0 in H.
apply eval_andlist'2andlist_semantic; auto.
Qed.

(**********************************)

(* orlist: disjonction of andlist *)

Definition orlist := list andlist.

Fixpoint orlist_semantic (l: orlist) : exprBZ:=
 match l with
   nil => !Z true_b
   | hd::tl => (andlist_semantic hd) |||Z (orlist_semantic tl)
 end.

(* Translation from a boolean expression to a orlist *)

Fixpoint dnf_expr_b2orlist (b: exprBZ) : option orlist :=
   match b with
      b1 |||Z b2 => option_app (dnf_expr_b2orlist b1) (dnf_expr_b2orlist b2)
      | _ => match (dnf_expr_b2andlist b) with
                     None => None
                     | Some b' => Some (b'::nil)
                 end
   end.

(* tool lemmas *)

Lemma dnf_expr_b2orlist_notempty: forall b b',
 dnf_expr_b2orlist b = Some b' -> 
 b' <> nil.
induction b; intros; simpl in H; try discriminate.
injection H; intros; subst b'; clear H.
red; intros; discriminate.
destruct (dnf_expr_b2orlist b1); destruct (dnf_expr_b2orlist b2); simpl in H; try discriminate.
injection H; intros; subst b'; clear H.
generalize (IHb1 o); intros.
intuition.
destruct o; intuition; discriminate.
generalize (dnf_expr_b2andlist_notempty b1); generalize (dnf_expr_b2andlist_notempty b2); intros.
destruct (dnf_expr_b2andlist b1); destruct (dnf_expr_b2andlist b2); simpl in H; try discriminate.
injection H; intros; subst b'; clear H.
generalize (H1 a); intros.
intuition.
destruct a; intuition; discriminate.
Qed.

Lemma orlist_semantic_app: forall l1 l2 s,
 l1 <> nil ->
 l2 <> nil ->
 eval_exprBZ (orlist_semantic (l1 ++ l2)) s =
 eval_exprBZ ((orlist_semantic l1) |||Z (orlist_semantic l2)) s.
induction l1.
intuition.
intros.
destruct l1; simpl orlist_semantic.
destruct l2; try tauto; simpl andlist_semantic; simpl orlist_semantic.
apply exprBZ_eq; split; intros; Omega_exprb.
destruct l2; try tauto; simpl andlist_semantic; simpl orlist_semantic.
simpl orlist_semantic in IHl1.
assert (a0::l1 <> nil); [red;  intro; discriminate | idtac].
generalize (IHl1 (a1::l2) s H1 H0); intros.
generalize (exprBZ_eq' _ _ _ H2); clear H2; intros.
inversion_clear H2.
eapply exprBZ_eq; split; intros.
clear H4; Eval_b_hyp_clean; inversion_clear H4; Eval_b_goal ; intuition.
assert (A1 : s |bZ= (andlist_semantic a0 |||Z orlist_semantic (l1 ++ a1 :: l2))); [Omega_exprb | generalize (H3 A1); clear H3 A1; intros; Omega_exprb].
clear H3.
simpl orlist_semantic in H4.
Eval_b_hyp_clean; Eval_b_goal.
assert ( exprBZ_sem (andlist_semantic a) s \/ (exprBZ_sem (andlist_semantic a0) s \/ (exprBZ_sem (orlist_semantic l1) s) \/ exprBZ_sem (andlist_semantic a1) s \/ exprBZ_sem (orlist_semantic l2) s) ).
intuition.
inversion_clear H2; [intuition |
assert (A1: s |bZ= ((andlist_semantic a0 |||Z orlist_semantic l1) |||Z (andlist_semantic a1 |||Z orlist_semantic l2))); [Omega_exprb | generalize (H4 A1); clear H4 A1; intros; Omega_exprb]
].
Qed.

(* if possible, the translation preserves the evaluation *)

Lemma dnf_expr_b2orlist_correct: forall b b',
 dnf_expr_b2orlist b = Some b' ->
 (forall s, eval_exprBZ b s = eval_exprBZ (orlist_semantic b') s).
induction b; intros; simpl in H; try discriminate.
(*>>=Z*)
injection H; intros; subst b'.
simpl orlist_semantic.
apply exprBZ_eq; split; intros; Eval_b_hyp_clean; Eval_b_goal; intuition.
(* \/ *)
generalize (dnf_expr_b2orlist_notempty b1); generalize (dnf_expr_b2orlist_notempty b2); intros.
destruct (dnf_expr_b2orlist b1); destruct (dnf_expr_b2orlist b2); simpl in H; try discriminate.
injection H; intros; subst b'; clear H.
rewrite orlist_semantic_app; [idtac | apply H1; auto | apply H0; auto].
simpl; rewrite (IHb1 o); auto; rewrite (IHb2 o0); auto.
(* /\ *)
generalize (dnf_expr_b2andlist_notempty b1); generalize (dnf_expr_b2andlist_notempty b2); intros.
generalize (dnf_expr_bool2andlist_correct b1); generalize (dnf_expr_bool2andlist_correct b2); intros.
destruct (dnf_expr_b2andlist b1); destruct (dnf_expr_b2andlist b2); simpl in H; try discriminate.
injection H; intros; subst b'; clear H.
simpl orlist_semantic.
apply exprBZ_eq; split; intros.
cut (s |bZ= andlist_semantic (a ++ a0)); [intros; Eval_b_hyp_clean; Eval_b_goal; intuition | idtac].
rewrite andlist_semantic_app.
simpl; simpl in H; rewrite <- (H2 a0); [rewrite <- (H3 a); auto | auto].
assert (s |bZ= andlist_semantic (a ++ a0)); [intros; Eval_b_hyp_clean; Eval_b_goal; intuition | idtac].
rewrite andlist_semantic_app in H4.
simpl; simpl in H4; rewrite <- (H2 a0) in H4; [rewrite <- (H3 a) in H4; auto | auto].
Qed.

(* simplification of all the constraints of the orlist *)

Fixpoint simpl_orlist (l: orlist): orlist :=
   match l with
      nil => nil
      | hd::tl => (simpl_andlist hd)::(simpl_orlist tl)
   end.

(* simplification preserves evaluation *)

Lemma simpl_orlist_correct: forall l ,
 (forall s, eval_exprBZ (orlist_semantic l) s =   eval_exprBZ (orlist_semantic (simpl_orlist l)) s).
induction l.
auto.
simpl orlist_semantic.
intros.
apply exprBZ_eq; split; intros.
assert (
(s |bZ= andlist_semantic a) \/
(s |bZ= orlist_semantic l)
).
Eval_b_hyp_clean.
inversion_clear H0.
left; Omega_exprb.
right; Omega_exprb.
inversion_clear H0.
rewrite simpl_andlist_correct in H1; Omega_exprb.
rewrite IHl in H1; Omega_exprb.
assert (
(s |bZ= andlist_semantic (simpl_andlist a)) \/
(s |bZ= orlist_semantic (simpl_orlist l))
).
Eval_b_hyp_clean.
inversion_clear H0.
left; Omega_exprb.
right; Omega_exprb.
inversion_clear H0.
rewrite <- simpl_andlist_correct in H1; Omega_exprb.
rewrite <- IHl in H1; Omega_exprb.
Qed.

(* elimination of the variable v in the orlist  *)

Fixpoint elim_var_orlist (l: orlist) (v: nat) {struct l} : orlist :=
   match l with
      nil => nil
      | hd::tl => (elim_var_andlist hd nil v)::(elim_var_orlist tl v)
   end.

(*   *)

Lemma elim_var_orlist_correct: forall l v,
  (forall s, eval_exprBZ (orlist_semantic l) s = true -> 
               eval_exprBZ (orlist_semantic (elim_var_orlist l v)) s = true ).
induction l.
simpl.
intuition.
simpl orlist_semantic.
intros.
assert (
(s |bZ= andlist_semantic a) \/
(s |bZ= orlist_semantic l)
).
Eval_b_hyp_clean.
inversion_clear H0.
left; Omega_exprb.
right; Omega_exprb.
cut(
(s |bZ= andlist_semantic (elim_var_andlist a nil v)) \/
(s |bZ= orlist_semantic (elim_var_orlist l v))
).
intros.
inversion_clear H1; Omega_exprb.
inversion_clear H0.
left.
apply elim_var_andlist_correct.
rewrite <- app_nil_end.
auto.
right.
apply IHl.
auto.
Qed.

(* Elimination of a list of variables *)

Fixpoint elim_varlist_orlist (l: orlist) (v: list nat) {struct v} : orlist :=
   match v with
      nil => l
      | hd::tl => (elim_varlist_orlist (elim_var_orlist l hd) tl)
   end.

(* *)

Lemma elim_varlist_orlist_correct: forall v l,
 (forall s, eval_exprBZ (orlist_semantic l) s = true ->
               eval_exprBZ (orlist_semantic (elim_varlist_orlist l v)) s = true).
induction v.
auto.
simpl elim_varlist_orlist.
intros.
apply IHv.
apply elim_var_orlist_correct.
auto.
Qed.

(* Evaluation of a non empty orlist without environment *)

Fixpoint eval_orlist' (o: orlist) : option bool :=
   match o with
     nil => Some false
     | hd::tl => orb_option (eval_andlist hd) (eval_orlist' tl)
   end.

(*  if possible, evaluation holds for every environment *)

Lemma eval_orlist'2orlist_semantic: forall a b,
  eval_orlist' a = Some b ->
  (forall s, eval_exprBZ (orlist_semantic a) s = b).
induction a; simpl; intros.
injection H; auto.
generalize ( eval_andlist2andlist_semantic a); intros.
destruct (eval_andlist a); simpl in H; try discriminate.
rewrite (H0 b0); auto.
destruct (eval_orlist' a0); simpl in H; try discriminate.
simpl in H.
lapply (IHa b1); intros.
rewrite H1.
injection H; auto.
auto.
Qed.

(* evaluation of an orlist without environement *)

Definition eval_orlist (a: orlist) : option bool :=
  if (beq_nat (length a) 0) then
      None 
  else     
      eval_orlist' a.

(*  if possible, evaluation holds for every environment *)

Lemma eval_orlist2orlist_semantic: forall a b,
  eval_orlist a = Some b ->
  (forall s, eval_exprBZ (orlist_semantic a) s = b).
intros.
unfold eval_orlist in H.
generalize (beq_nat_classic (length a) 0); intro X; inversion_clear X.
rewrite H0 in H; discriminate.
rewrite H0 in H.
apply eval_orlist'2orlist_semantic; auto.
Qed.

(**************************)

(* Translation from a boolean expression to an orlist  *)

Definition expr_b2constraints (b: exprBZ) : option orlist :=
      dnf_expr_b2orlist (dnf_expr_b b).

(* if possible, translation preserves evaluation  *)

Lemma expr_b2constraints_correct: forall b x,
  expr_b2constraints b = Some x ->
  (forall s, eval_exprBZ b s = eval_exprBZ (orlist_semantic x) s).
intros.
unfold expr_b2constraints in H.
rewrite (dnf_expr_b_correct b s).
rewrite (dnf_expr_b2orlist_correct _ _ H).
auto.
Qed.

(**************************)

(* Fourrier-Motzkin ???  *)

(* The algorithm is

  i) put the boolean expression in normale form
  ii) elimine all its variables
  iii) evaluate wihtout environment
  iv) return the negation 

*)

Definition presburgerZ (b: exprBZ) : bool :=
   match (expr_b2constraints (!Z b)) with
      None => false
      | Some ol => match (eval_orlist (simpl_orlist (elim_varlist_orlist ol (exprBZ_var b)))) with
                               None => false
                               | Some res => negb res
                            end
  end.

(* if the function result is true then the original expression is true for every environment  *)

Lemma presburgerZ_correct: forall b,
   presburgerZ b = true ->
   (forall s, s |bZ= b).
intros.
unfold presburgerZ in H.
assert (
(s |bZ= b) \/
(s |b!Z= b)
).
destruct eval_exprBZ.
left; auto.
right; auto.
inversion_clear H0; auto.
generalize (expr_b2constraints_correct (!Zb)); intros.
destruct (expr_b2constraints (!Zb)).
assert (
s |bZ= !Zb
).
Omega_exprb.
clear H1.
rewrite (H0 o) in H2; auto.
clear H0.
assert (
eval_orlist (simpl_orlist (elim_varlist_orlist o (exprBZ_var b))) = None \/
exists v, eval_orlist (simpl_orlist (elim_varlist_orlist o (exprBZ_var b))) = Some v
).
destruct (eval_orlist (simpl_orlist (elim_varlist_orlist o (exprBZ_var b)))).
right; exists b0; auto.
left; auto.
inversion_clear H0.
rewrite H1 in H; inversion H.
inversion_clear H1.
rewrite H0 in H.
destruct x.
simpl in H; discriminate.
generalize (eval_orlist2orlist_semantic _ _ H0 s); intros.
generalize (elim_varlist_orlist_correct (exprBZ_var b) o s H2); intros.
rewrite  (simpl_orlist_correct) in H3.
Omega_exprb.
inversion H.
Qed.

(*

Example

Definition A:= (((var_eZ 0) +eZ (var_eZ 1)) *eZ ((var_eZ 0) +eZ (var_eZ 1))).
Definition B:= (((var_eZ 0) *eZ (var_eZ 0)) +eZ ((var_eZ 1) *eZ (var_eZ 1)) +eZ ((int_eZ 2) *eZ (var_eZ 0) *eZ (var_eZ 1))).

Definition test1 : exprBZ :=
  (A ==Z B).
Eval compute in (presburgerZ test1).

Definition t15 := ((var_eZ 1 >>Z var_eZ 2) =bZ> ((var_eZ 2 >>Z var_eZ 0) =bZ> (var_eZ 1 >>Z var_eZ 0))).
Eval compute in (presburgerZ t15).

*)


(***********************************)

Lemma new_cut: forall P Q,
  P -> (P -> Q) -> Q.
  intros; tauto.
Qed.

Ltac new_cut P :=
  match goal with
    | |- ?Q =>
      apply (new_cut P Q)
  end.


(* FrontEnd Tactics for the presburgerZ function  *)

(* Determine if t is a Coq positive variable *)

Ltac Is_pos_var p :=
  match p with
    xH => false
    | xI ?n => false
    | xO ?n => false
    | _ => true
  end.
    


(* Determine if t is a Coq Z variable *)

Ltac Is_Z_var t :=
  match t with
    | Z0 => false
    | Zpos ?e  => (Is_pos_var e)
    | Zneg ?e => (Is_pos_var e)
    | _ => true
  end.

(* add an element to a list only if not already present *)

Ltac Add_list e l :=
  match l with

    | e::?tl => l

    | ?hd::?tl => 
      let x:= (Add_list e tl) in (
        constr:(hd::x)
      )      

    | (@nil Z) => 
      constr:(e::nil)

  end.

(*Concat tweo lists using previous tactic *)

Ltac Concat_list l1 l2 :=
  match l1 with
    | ?hd::?tl => 
      let x:= (Add_list hd l2) in (
        Concat_list tl x
      )
    | nil => l2
  end.      

(* Build the list of the Coq variable inside the term t *)
(* Such a list is used as an environment  *)
Ltac Build_env t :=
  match t with
    | (?t1 + ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 - ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 * ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 = ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 -> ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 > ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 < ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 >= ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 <= ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 /\ ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 \/ ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (?t1 <> ?t2)%Z => 
      let x:= (Build_env t1) in (
        let y:= (Build_env t2) in (
          Concat_list x y
        )
      )
    | (~ ?t1)%Z => 
      let x:= (Build_env t1) in (
        x
      )
    | _ => 
      let x:=(Is_Z_var t) in (
        match (eval compute in x) with
          true => constr:(t::nil)
          | false => constr:(@nil Z)
        end
        )
  end.

(* Find the indice c of variable v in l *)

Ltac Find_var v l c :=
  match l with

    | v::?tl => 
      constr:c

    | ?hd::?tl => 
      let x:=(eval compute in (S c)) in (Find_var v tl x)

    | (@nil Z) => 
      constr:0

  end.

(* return the index of variable v in list l *)

Ltac Get_var_index v l := Find_var v l O.

(* Debug Tactic *)

Ltac Print t := assert (t = t).

(* Translate a Coq term to an arithmetic expression using a list of variables l *)

Ltac To_exprZ t l :=
  match t with
    | (?t1 + ?t2)%Z => 
      let x:= (To_exprZ t1 l) in (
        let y:= (To_exprZ t2 l) in (
          constr:(x +eZ y)
        )
      )
    | (?t1 - ?t2)%Z => 
      let x:= (To_exprZ t1 l) in (
        let y:= (To_exprZ t2 l) in (
          constr:(x -eZ y)
        )
      )
    | (?t1 * ?t2)%Z => 
      let x:= (To_exprZ t1 l) in (
        let y:= (To_exprZ t2 l) in (
          constr:(x *eZ y)
        )
      )
    | _ => 
      let x:=(Is_Z_var t) in (
        match (eval compute in x) with
          | true => 
            let y:= (Get_var_index t l) in 
             constr:(var_eZ y)
          | false =>               
             constr:(int_eZ t)
        end
        )
  end.

(* Translate a Coq term to a boolean expression using a list of variables l *)

Ltac To_exprBZ t l :=
  match t with
    | (?t1 = ?t2) =>
      let x := (To_exprZ t1 l) in (
        let y := (To_exprZ t2 l) in (
          constr:(x ==Z y)
        )
      )
    | (?t1 > ?t2)%Z =>
      let x := (To_exprZ t1 l) in (
        let y := (To_exprZ t2 l) in (
          constr:(x >>Z y)
        )
      )
    | (?t1 >= ?t2)%Z =>
      let x := (To_exprZ t1 l) in (
        let y := (To_exprZ t2 l) in (
          constr:(x >>=Z y)
        )
      )
    | (?t1 < ?t2)%Z =>
      let x := (To_exprZ t1 l) in (
        let y := (To_exprZ t2 l) in (
          constr:(x <<Z y)
        )
      )
    | (?t1 <= ?t2)%Z =>
      let x := (To_exprZ t1 l) in (
        let y := (To_exprZ t2 l) in (
          constr:(x <<=Z y)
        )
      )
    | (?t1 -> ?t2) =>
      let x := (To_exprBZ t1 l) in (
        let y := (To_exprBZ t2 l) in (
          constr:(x =bZ> y)
        )
      )
    | (?t1 /\ ?t2) =>
      let x := (To_exprBZ t1 l) in (
        let y := (To_exprBZ t2 l) in (
          constr:(x &&&Z y)
        )
      )
    | (?t1 \/ ?t2) =>
      let x := (To_exprBZ t1 l) in (
        let y := (To_exprBZ t2 l) in (
          constr:(x |||Z y)
        )
      )
    | (?t1 <> ?t2) =>
      let x := (To_exprBZ t1 l) in (
        let y := (To_exprBZ t2 l) in (
          constr:(x =/=Z y)
        )
      )
    | (~ ?t1) =>
      let x := (To_exprBZ t1 l) in (
          constr:(!Zx)
      )
  end.

(* return true if the v element is in l *)

Ltac In_list v l :=
  match l with

    | v::?tl =>
      true

    | ?hd::?tl =>
      In_list v tl

    | (@nil Z) =>
      false

  end.

(* return true if at least one element of l1 is in l2  *)

Ltac In_list_list l1 l2 :=
  match l1 with

    | ?hd::?tl =>
      let x:=(In_list hd l2) in
        match (eval compute in x) with
          | true => true
          | false => (In_list_list tl l2)
        end

    | (@nil Z) =>
      false

  end.

(*
Ltac PresburgerZ_decision := 
  (* Introduction of all pertinent hypothesis *)
  match goal with
    | |- ?G =>
      (* l is the environment including all coq variable of the goal *)
      let l := (Build_env G) in (
        (* x is the boolean expression corresponding to the goal with respect to the environement l *)
        let x := (To_exprBZ G l) in (
          (* a proof of the goal can be achieve by proving that evaluation of x for l is true  *)
          cut (eval_exprBZ x l = true); [
            (* we prove entailement  *)
            let h:= fresh in (
              intro h; let y := fresh in (
                generalize (proj1 (exprBZ_semantic_good x l) h); intro y; simpl in y; intros; tauto
              )
            )| 
            (* To prove the evaluation we use the persburgerZ function  *)
            eapply presburgerZ_correct; simpl; apply refl_equal]
        )
      )
  end.
*)


(* Presburger without any seeking to hypothesis *)
Ltac PresburgerZ_decision := 
  (* Introduction of all pertinent hypothesis *)
  match goal with
    | |- ?G =>
      (* l is the environment including all coq variable of the goal *)
      let l := (Build_env G) in (
        (* x is the boolean expression corresponding to the goal with respect to the environement l *)
        let x := (To_exprBZ G l) in (
          (* a proof of the goal can be achieve by proving that evaluation of x for l is true  *)
          new_cut (eval_exprBZ x l = true); [
            (* To prove the evaluation we use the persburgerZ function  *)
            eapply presburgerZ_correct; simpl; apply refl_equal
              |
            (* we prove entailement  *)
            let h:= fresh in (
              intro h; let y := fresh in (
                generalize (proj1 (exprBZ_semantic_good x l) h); intro y; simpl in y; intros; tauto
              )
            )]            
        )
      )
  end.


(* generalize and clear the first hypothesis that comport a variable in l and then call PresburgerZ or PresburgerZ_decision*)
Ltac Intro_hyp_var_list l :=
  match goal with
    | id: ?a%Z |- _ =>
      let l' := (Build_env a) in (
        let x := (In_list_list l' l) in (
          match (eval compute in x) with
            | true => generalize id; clear id; (PresburgerZ || PresburgerZ_decision)
            | false => fail
          end
        )
      )

    | _ => PresburgerZ_decision
  end
(* call Intro_hyp_var_list with the list of Coq variables present in the goal  *)
with PresburgerZ :=
  match goal with
    | |- ?G =>
      let l := (Build_env G) in (
        Intro_hyp_var_list l
      )
  end.    


(* Try to find a hypothesis that can be prove wrong *)
Ltac ContradictionZ :=
  match goal with
    | id: ?A |- _ =>
      assert (~A); [PresburgerZ | tauto]
    | _ => fail
  end.
    

Close Local Scope presburgerZ_scope.

(*

Open Local Scope Z_scope.


Lemma t: forall a b,
  (a < b \/ b < a) -> a <> b.
  intros.
  PresburgerZ.
Show Proof.
Qed.

Lemma t': forall a b c,
  (a > b /\ a < b) -> a=c.
  intros.
  ContradictionZ.
Qed.

Lemma t'':
  1 <= 1.
  PresburgerZ.
Qed.

Definition pp (n: nat) : Z :=
  Z_of_nat n * 66.

Lemma test_omega: forall a b c,
  pp a > b -> b > c -> pp a > c.
  intros.
  PresburgerZ.
  Show Proof.
Qed.

Lemma test_presburgerZ: forall a b c,
  a >= b -> b > c -> a > c.
  intros; PresburgerZ.
  Show Proof.
Qed.

Lemma test2_omega: forall a b c,
  a < b -> b < c -> a < c.
  intros; omega.
  Show Proof.
Qed.

Lemma test2_presburgerZ: forall a b c,
  a < b -> b < c -> a < c.
  intros; PresburgerZ.
  Show Proof.
Qed.


Lemma t1_ring: forall a b,
  (a + b) * (a + b) = a*a + b*b + 2*a*b.
  intros.
  ring.
  Show Proof.
Qed.  

Lemma t1_presburgerZ: forall a b,
  (a + b) * (a + b) = a*a + b*b + 2*a*b.
  intros; PresburgerZ.
  Show Proof.
Qed.  

Close Local Scope Z_scope.

*)











